#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "http.h"


/* Create a TCP connection to host and port. Returns a file
* descriptor on success, -1 on error. */
int
tcpconnect (const char *host, int port)
{
	struct hostent *h;
	struct sockaddr_in sa;
	int s;

	h = gethostbyname (host);
	if (!h || h->h_length != sizeof (struct in_addr)) {
		fprintf (stderr, "%s: no such host\n", host);
		return -1;
	}

	/* Create a TCP socket. */
	s = socket (AF_INET, SOCK_STREAM, 0);
	
	memset (&sa, 0, sizeof (sa));
	sa.sin_family = AF_INET;
	sa.sin_port = htons (0); 
	sa.sin_addr.s_addr = htonl (INADDR_ANY); 
	
	/* Use bind to set an address and port number for our end of the
	* TCP connection. */
	if (bind (s, (struct sockaddr *) &sa, sizeof (sa)) < 0) {
		perror ("bind");
		close (s);
		return -1;
	}

	/* Now use h to set set the destination address. */
	sa.sin_port = htons (port);
	sa.sin_addr = *(struct in_addr *) h->h_addr;
	/* And connect to the server */
	if (connect (s, (struct sockaddr *) &sa, sizeof (sa)) < 0) {
		perror (host);
		close (s);
		return -1;
	}
	
	return s;
}


void
synchronous_proxy (int s)
{
	httpreq hr;
	strbuf sbuf;
	int ret=0,s2,nread;

	while ( (0 >= sbuf.tosuio()->input(s)) && errno == EAGAIN)
		if ( 0 != (ret=httpreq_parse (&hr, sbuf.tosuio())) )
			break;
	if ( (0 < ret) && (0 < (s2=tcpconnect(hr.host(),hr.port()))) ){
		//dump the whole request out to the newly opened conn'x
		if (0 < hr._headers.tosuio()->output(s2)){
			//and just copy read stuff to old socket (4k=page size)
			char buf[4096];
			while ( 0 < (nread=read(s2,buf,sizeof(buf))) )
				write(s,buf,nread);
		}
	}
}

void
do_proxy (int c)
{
  synchronous_proxy (c);
  close (c);
}

int
tcplisten (int port)
{
  // new socket
  int s = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (s < 0) {
    perror ("socket");
    exit (1);
  }

  // don't whine about address already in use
  int n = 1;
  if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof (n)) < 0) {
    perror ("setsockopt");
    exit (1);
  }

  // bind socket to some port
  sockaddr_in sin;
  bzero (&sin, sizeof (sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons (port);
  sin.sin_addr.s_addr = INADDR_ANY;
  if (bind (s, (sockaddr *) & sin, sizeof (sin)) < 0) {
    perror ("bind");
    exit (1);
  }

  // listen for incoming connections
  if (listen (s, 5) < 0) {
    perror ("listen");
    exit (1);
  }

  return s;
}

int
main (int argc, char **argv)
{
  int port;
  if (argc != 2 || !(port = atoi (argv[1])) || port >= 65536) {
    fprintf (stderr, "Usage: %s PORT\n", argv[0]);
    exit (1);
  }

  // liten for TCP connections on port
  int s = tcplisten (port);

  for (;;) {
    sockaddr_in clisin;
    socklen_t clisinlen = sizeof (clisin);
    bzero (&clisin, clisinlen);
    int c = accept (s, (sockaddr *) & clisin, &clisinlen);
    if (c < 0) {
      perror ("accept");
      exit (1);
    }

    do_proxy (c);		// Perform web proxy functionality
  }
}
