#include "webproxy2.h"

int main(int argc, char ** argv)
{
	async_init();	
	//check for port in command line
	int port;
	if(argc == 2) 
		port = atoi(argv[1]);
	else
		port = DEF_PORT;

	warn << "Async WebProxy v0.1 listening on " << port << "\n";

	//start listening
	int fd = inetsocket(SOCK_STREAM, port);
	if (fd < 0)
          fatal << "inetsocket\n";
     
	make_async(fd);
	tcp_nodelay(fd);

     if (listen(fd, MAX_CONN) < 0)
       fatal << "listen\n";
   
     fdcb(fd, selread, wrap(cs_accept_connection, fd));
   
     amain();

}

void
cs_accept_connection(int fd)
{
	
       struct sockaddr_in sin;
       unsigned sinlen = sizeof(sin);
   
       int cs = accept(fd, (struct sockaddr *) &sin, &sinlen);
       if (cs >= 0) {
         	warn << "----------->Accepted connection on: " << cs << "\n";
       } else if (errno != EAGAIN){
     	  warn << "Accept; errno = " << errno << "\n";
     	  return;
       }
	make_async(cs);
	tcp_nodelay(cs);

	connx *cx = New(connx);
	cx->cs = cs;
	fdcb(cs, selread,wrap(cs_read_ready, cx) );
}

void
cs_read_ready(connx *cx)
{
	
	int n = cx->buf.tosuio()->input(cx->cs);
//	warn << "cs: read: " << n << "\n"; 

	if( (n < 0) && (errno != EAGAIN)) {
    	warn << "cs("<<cx->cs<<"): error in read:" << errno << " , closing connection...\n";
   		fdcb(cx->cs, selread, 0);
		delete cx;
   		return;
   	}
   	
   	//call parser and check if request is complete
   	int ret = cx->req.parse(cx->buf.tosuio());
   	switch (ret){
   		case -1:
   			warn << "cs("<<cx->cs<<"): error in parse, closing connection...\n";
	   		fdcb(cx->cs, selread, 0);
			delete cx;
   			return;
   		case 0:
   			warn << "cs("<<cx->cs<<"): parser needs more data, waiting...\n";
   			//fdcb(cx->cs, selread, wrap(cs_read_ready, cx));
 			return;
   		case 1:
   			warn << "cs("<<cx->cs<<"): parsing complete, requesting " << cx->req._path << " from "
   				<< cx->req.host() << ":" << cx->req.port() << "\n";
   			fdcb(cx->cs, selread, 0);
   		    
   		    //check cache
   		    cacheEntry *ceT;
   		    bool getNew = true;
   		    
			type_Cache::iterator iter = gCache.find(cx->req.urlhash());
			
			if( iter != gCache.end() )
			{	//found!
			    warn << "cs("<<cx->cs<<"): Cache Hit!\n";
  				ceT = (*iter).second;
  				time_t now = time (NULL);
  				
  				//check staleness
  				if ( (ceT->expires>0) && (ceT->expires <= now) ){
  					warn << "cs("<<cx->cs<<"): Cache Entry Stale (expires: "<<ceT->expires<<"), requesting anew\n";
  				}else if ( (cx->req.if_modified_since>0) && (ceT->last_modified>0) &&
  						cx->req.if_modified_since > ceT->last_modified){
  					//check if we should send a 304
  					strbuf send;
  					send << "HTTP/1.0 304 Not modified\r\n\r\n";
  					send.tosuio()->output(cx->cs);
  					warn << "cs("<<cx->cs<<"): Not modified, returning 304\n";
					getNew = false;
				}else{
					//entry is valid and not stale -> send
					cx->ce = ceT;
					cx->cs_send_from_cache = true;
					warn << "cs("<<cx->cs<<"): Cache Entry valid, sending...\n";
  					fdcb(cx->cs, selwrite, wrap(cs_write_ready, cx));
					getNew = false;
				}
  			}//end if (found)
			
			if (getNew){
   		    	warn << "cs("<<cx->cs<<"): " << "No cache hit, requesting...\n";
				cx->ce = new(cacheEntry);
   		    	//open connection
	  		    cx->buf.tosuio()->clear();
				tcpconnect(cx->req.host(), cx->req.port(), wrap(ss_connected, cx));
   			}
	}//end switch
//	warn << "returning\n";
}//end cs_read_ready

void
ss_connected(connx *cx, int newSock)
{
	if(newSock < 0){
        	warn << "ss("<<cx->ss<<"): tcpconnect failed, closing connection...\n";
		delete cx;
   		return;
   	}
     
	cx->ss = newSock;

	fdcb(cx->ss, selwrite, wrap(ss_write_ready, cx));
	
}
	
void
ss_write_ready(connx *cx){
	//add newline
	warn << "request is: " << cx->req._headers;
	cx->req._headers << "\r\n";
	//send request through
	int ret = cx->req._headers.tosuio()->output(cx->ss);

	switch (ret){
		case -1:
			warn << "ss("<<cx->ss<<"): sending headers failed, closing connection...\n";
	   		fdcb(cx->ss, selwrite, 0);
			delete cx;
   			return;
   		case 0:
   			//call again
   			return;
   		case 1:
   			fdcb(cx->ss, selwrite, 0);
   		   	//everything went fine... wait for response
		   	fdcb(cx->ss, selread, wrap(ss_read_ready, cx));
   			//else: call again
   	}
}
    
void
ss_read_ready(connx *cx){
    
    int n = cx->buf.tosuio()->input(cx->ss);
   
	if(n < 0){
    		warn << "ss("<<cx->ss<<"): error in read, closing connection...\n";
   		fdcb(cx->ss, selread, 0);
		delete cx;
   	}else if (n==0){	
   		fdcb(cx->ss, selread, 0);
	}else{
   	//call parser and check if request is complete
		if (cx->cs_sent_headers)
   			fdcb(cx->cs, selwrite, wrap(cs_write_ready, cx));
		else{
				
   			int ret = cx->resp.parse(cx->buf.tosuio());
   			switch (ret){
   			case -1:
   				warn << "ss("<<cx->ss<<"): error in response, closing connection...\n";
	   			fdcb(cx->ss, selread, 0);
				delete cx;
   				return;
   			case 0:
   				warn << "ss("<<cx->ss<<"): response parser needs more data, waiting...\n";
   				return;
   			case 1:
   				warn << "ss("<<cx->ss<<"): response complete, forwarding...\n";
   				fdcb(cx->cs, selwrite, wrap(cs_write_ready, cx));
			}
		}
	} 			
}

void
cs_write_ready(connx *cx){
	char buf[2];
	int ret = 0, size = 0;
	
	//send from cache?
	if (cx->cs_send_from_cache){
		strbuf h;
		strbuf c;
		if (!cx->cs_sent_headers){	
			//copy over to temp strbufs
			h << cx->ce->had;
			c << cx->ce->con;
			//and send headers
			size = h.tosuio()->resid(); 
			if (0 > h.tosuio() -> output(cx->cs)){
				warn << "cs("<<cx->cs<<"): cached header send failed, closing connection...\n";
		   		fdcb(cx->cs, selwrite, 0);
				delete cx;
   				return;
			}
			strcpy(buf,"\r\n");
			write(cx->cs,buf,2);
			cx->fwed += size+2;
			cx->cs_sent_headers = true;
//			warn << "cs("<<cx->cs<<"): cached header sent: " << size << "bytes \n";

		}
		//and content
		size = c.tosuio()->resid(); 
		ret  = c.tosuio()->output(cx->cs);
		switch (ret){
			case -1:
				warn << "cs("<<cx->cs<<"): sending from cache failed, closing connection...\n";
	   			fdcb(cx->cs, selwrite, 0);
				delete cx;
   				break;
   			case 0:
   				//call again
	   			break;
			case 1:
				cx->fwed += size;
//				warn << "cs("<<cx->cs<<"): cached content sent: " << size << "bytes \n";
   				//if all sent: unregister
   				if (0 == c.tosuio()->resid()){
   					fdcb(cx->cs, selwrite, 0);
					warn << "cs("<<cx->cs<<"): cached sent : " << cx->fwed << " bytes \n";
					delete cx;
				}
			}
	}else{
		//headers sent?
		if (!cx->cs_sent_headers){	
			warn << "cs("<<cx->cs<<"): Header is: " << cx->resp._headers << "\n";
		
			//before sending copy over to cacheEntry
			cx->ce->had << cx->resp._headers;
			 
			size = cx->resp._headers.tosuio()->resid(); 
			if (0 > cx->resp._headers.tosuio()->output(cx->cs)){
				warn << "cs("<<cx->cs<<"): sending failed, closing connection...\n";
		   		fdcb(cx->cs, selwrite, 0);
				delete cx;
   				return;
			}
			strcpy(buf,"\r\n");
			write(cx->cs,buf,2);
			cx->fwed += size+2;
//			warn << "cs("<<cx->cs<<"): header sent: " << size << "bytes \n";

			cx->cs_sent_headers = true;
		}

		//copy content over to cacheentry, also
		cx->ce->con << cx->buf;
		//save size
		size = cx->buf.tosuio()->resid(); 
		ret  = cx->buf.tosuio()->output(cx->cs);

		switch (ret){
			case -1:
				warn << "cs("<<cx->cs<<"): sending failed, closing connection...\n";
	   			fdcb(cx->cs, selwrite, 0);
				delete cx;
   				break;
   			case 0:
   				//call again
	   			break;
   			case 1:
				cx->fwed += size;
		//		warn << "cs("<<cx->cs<<"): content sent: " << size << "bytes \n";
   				//unregister, until more to read
   				fdcb(cx->cs, selwrite, 0);
				
				//and if all read close:
				int f = recv(cx->ss,buf,1,MSG_PEEK); 
			//	warn << "cs("<<cx->cs<<"): msg_peek is: " << f << ",err: " << errno << "\n";
				if (f==0){
					//request is complete and all fw'ed:
					//go ahead unregister callbacks and add to cache if appr.
		   			fdcb(cx->cs, selwrite, 0);
   					fdcb(cx->ss, selread, 0);
					warn << "cs("<<cx->cs<<"): sent : " << cx->fwed << " bytes \n";
					//and add entry to cache if conditions are met
					if (!cx->resp.nocache && !cx->req.nocache && !cx->req.authorization){
						warn << "cs("<<cx->cs<<"): caching " << cx->req.url() << " (" << cx->fwed << " bytes) \n";
						cx->ce->date = cx->resp.date;
						cx->ce->expires = cx->resp.expires;
						cx->ce->last_modified = cx->resp.last_modified;
					
						gCache[cx->req.urlhash()] = cx->ce;
					}
					delete cx;
				}//end if(f==0)
  		}//end switch
   	}//end else
}//end cs_write_ready
